
#include "kDist.h"

FILE *g_debugFilePtr;

DistEntry *distEntry_new(float dist, equivalenceTuple **eqTuple1P, equivalenceTuple **eqTuple2P)
{
	DistEntry *returnValue;

	returnValue = (DistEntry *) malloc(sizeof(DistEntry));

	returnValue->dist = dist;
	returnValue->hashSlot = NULL;
	returnValue->hashSlotNo = 0;
	returnValue->distSlot = NULL;
	returnValue->distSlotNo = 0;


	// sort the eq tuple by ID
	if ((*eqTuple1P)->id < (*eqTuple2P)->id)
	{
		returnValue->eqTuple1 = (*eqTuple1P);
		returnValue->eqTuple2 = (*eqTuple2P);
	}
	else
	{
		returnValue->eqTuple1 = (*eqTuple2P);
		returnValue->eqTuple2 = (*eqTuple1P);
	}

	return returnValue;
}


Hash *hash_new(int noOfSlot)
{
	Hash *returnValue;

	returnValue = (Hash *) malloc(sizeof(Hash));

	returnValue->noOfSlot = noOfSlot;
	returnValue->slot = (HashSlot *) malloc(sizeof(HashSlot)*noOfSlot);
	memset(returnValue->slot, 0, sizeof(HashSlot)*noOfSlot);
	

	return returnValue;
}

void kDist_insertDistEntry2HashSlot(HashSlot **hashSlotP, DistEntry **distEntryP, int slotNo)
{
	int noOfEntry;

	DistEntry *distEntry;
	HashSlot *hashSlot;

	hashSlot = (*hashSlotP);

	noOfEntry = hashSlot->noOfEntry;

	if (noOfEntry == 0)
	{
		hashSlot->entry = (DistEntry **) malloc(sizeof(DistEntry *));
	}
	else
	{
		hashSlot->entry = (DistEntry **) realloc(hashSlot->entry, sizeof(DistEntry *)*(noOfEntry+1));
	}

	hashSlot->entry[noOfEntry] = (*distEntryP);
	hashSlot->noOfEntry++;

	distEntry = (*distEntryP);

	distEntry->hashSlot = hashSlot;
	//distEntry->hashSlotNo = slotNo;
	distEntry->hashSlotNo = noOfEntry;
	
}

void kDist_insertDistEntry2Hash(Hash *hash, DistEntry **distEntryP)
{
	HashSlot *hashSlot;

	DistEntry *distEntry;

	int slotNo;
	int id1, id2;

	distEntry = (*distEntryP);

	id1 = distEntry->eqTuple1->id;
	id2 = distEntry->eqTuple2->id;

	// find the slot no
	slotNo = (id1 + id2)%hash->noOfSlot;

	// the hash slot
	hashSlot = &(hash->slot[slotNo]);

	// insert to the hashSlot
	kDist_insertDistEntry2HashSlot(&hashSlot, distEntryP, slotNo);
}

void kDist_removeEntryFromHashSlot(HashSlot *hashSlot, int slotNo)
{
	int i;
	int noOfEntry;

	DistEntry **newEntry;
	DistEntry **oldEntry;

	noOfEntry = hashSlot->noOfEntry;

	oldEntry = hashSlot->entry;

	if (noOfEntry == 1)
	{
		free(oldEntry);

		hashSlot->noOfEntry = 0;
	}
	else
	{
		newEntry = (DistEntry **) malloc(sizeof(DistEntry *)*(noOfEntry-1));

		if (slotNo > 0)
		{
			memcpy(newEntry, oldEntry, sizeof(DistEntry *)*(slotNo));
		}

		if (slotNo < noOfEntry-1)
		{
			memcpy(&(newEntry[slotNo]), &(oldEntry[slotNo+1]), sizeof(DistEntry **)*(noOfEntry-slotNo-1));
		}

		hashSlot->noOfEntry--;

		// update
		for (i = slotNo+1; i < noOfEntry; i++)
		{
			oldEntry[i]->hashSlotNo--;
		}

		hashSlot->entry = newEntry;

		free(oldEntry);
	}
}

void kDist_removeDistEntryFromHashSlot(DistEntry *distEntry)
{
	HashSlot *hashSlot;
	int slotNo;

	hashSlot = distEntry->hashSlot;
	slotNo = distEntry->hashSlotNo;

	kDist_removeEntryFromHashSlot(hashSlot, slotNo);
}


DistStorage *distStorage_new(float delta, float maxValue)
{
	DistStorage *returnValue;

	int noOfSlot;

	returnValue = (DistStorage *) malloc(sizeof(DistStorage));

	noOfSlot = (int) (maxValue/delta) + 2;
	returnValue->noOfSlot = noOfSlot;
	returnValue->delta = delta; 
	returnValue->slot = (DistSlot *) malloc(sizeof(DistSlot)*noOfSlot);
	memset(returnValue->slot, 0, sizeof(DistSlot)*noOfSlot);

	return returnValue;
}

DistSlot *distSlot_new()
{
	DistSlot *returnValue;

	returnValue = (DistSlot *) malloc(sizeof(DistSlot));

	returnValue->entry = NULL;
	returnValue->noOfEntry = 0;

	return returnValue;
}

void printDistStorage(FILE *filePtr, DistStorage *distStorage)
{
	int i, j;

	DistSlot *distSlot;

	DistEntry *distEntry;

	equivalenceTuple *eqTuple1;
	equivalenceTuple *eqTuple2;

	int id1, id2;

	for (i = 0; i < distStorage->noOfSlot; i++)
	{
		distSlot = &(distStorage->slot[i]);

		printf(" ~~~~~ Slot No. %d ~~~~~~\n", i);
		printf("    noOfEntry : %d\n", distSlot->noOfEntry);
		fprintf(filePtr, " ~~~~~ Slot No. %d ~~~~~~\n", i);
		fprintf(filePtr, "    noOfEntry : %d\n", distSlot->noOfEntry);
		
		for (j = 0; j < distSlot->noOfEntry; j++)
		{
			distEntry = distSlot->entry[j];

			eqTuple1 = distEntry->eqTuple1;
			eqTuple2 = distEntry->eqTuple2;

			id1 = eqTuple1->id;
			id2 = eqTuple2->id;

			printf("     %d. (%d, %d)\n", j , id1, id2);
			printf("         hashSlot: %p\n", distEntry->hashSlot);
			printf("         distSlot: %p\n", distEntry->distSlot);
			printf("         dist    : %.6f\n", distEntry->dist);
			printf("         hashSlotNo: %d\n", distEntry->hashSlotNo);
			printf("         distSlotNo: %d\n", distEntry->distSlotNo);
			fprintf(filePtr, "     %d. (%d, %d)\n", j , id1, id2);
			fprintf(filePtr, "         hashSlot: %p\n", distEntry->hashSlot);
			fprintf(filePtr, "         distSlot: %p\n", distEntry->distSlot);
			fprintf(filePtr, "         dist    : %.6f\n", distEntry->dist);
			fprintf(filePtr, "         hashSlotNo: %d\n", distEntry->hashSlotNo);
			fprintf(filePtr, "         distSlotNo: %d\n", distEntry->distSlotNo);
		}
	}
}


void kDist_insertDistEntry2DistSlot(DistSlot **distSlotP, DistEntry **distEntryP, int slotNo)
{
	int noOfEntry;
	DistEntry *distEntry;
	DistSlot *distSlot;

	distSlot = (*distSlotP);

	noOfEntry = distSlot->noOfEntry;

	if (noOfEntry == 0)
	{
		distSlot->entry = (DistEntry **) malloc(sizeof(DistEntry *));
	}
	else
	{
		distSlot->entry = (DistEntry **) realloc(distSlot->entry, sizeof(DistEntry *)*(noOfEntry+1));
	}

	distSlot->entry[noOfEntry] = (*distEntryP);
	distSlot->noOfEntry++;

	distEntry = (*distEntryP);

	distEntry->distSlot = distSlot;
	//distEntry->distSlotNo = slotNo;
	distEntry->distSlotNo = noOfEntry;
}

void kDist_insertDistEntry2DistStorage(DistStorage *distStorage, DistEntry **distEntryP)
{
	DistSlot *distSlot;
	DistSlot *aDistSlot;
	DistEntry *distEntry;

	int slotNo;
	float delta;
	float dist;

	int noOfSlot;

	distEntry = (*distEntryP);
	dist = distEntry->dist;

	delta = distStorage->delta;
	distSlot = distStorage->slot;
	noOfSlot = distStorage->noOfSlot;

	slotNo = (int) (dist/delta);

	if (slotNo >= noOfSlot)
	{
		printf("No enough dist slot!\n");
		exit(0);

		// need to reallocate the memory
//		if (noOfSlot == 0)
//		{
//			distSlot = (DistSlot *) malloc(sizeof(DistSlot)*(slotNo+1));
//		}
//		else
//		{
//			distSlot = (DistSlot *) realloc(distSlot, sizeof(DistSlot)*(slotNo+1));
//		}

//		memset(&(distSlot[noOfSlot]), 0, sizeof(DistSlot)*(slotNo+1-noOfSlot));

//		distStorage->noOfSlot = slotNo+1;
//		distStorage->slot = distSlot;
	}
	else
	{
		// no need to reallocat the memory
	}

	aDistSlot = &(distSlot[slotNo]);

	// insert to the distSlot
	kDist_insertDistEntry2DistSlot(&aDistSlot, distEntryP, slotNo);

}

void kDist_removeEntryFromDistSlot(DistSlot *distSlot, int slotNo)
{
	int i;
	int noOfEntry;

	DistEntry **newEntry;
	DistEntry **oldEntry;

	noOfEntry = distSlot->noOfEntry;

	oldEntry = distSlot->entry;

	if (noOfEntry == 1)
	{
		free(oldEntry);

		distSlot->noOfEntry = 0;
	}
	else
	{
		newEntry = (DistEntry **) malloc(sizeof(DistEntry *)*(noOfEntry-1));

		if (slotNo > 0)
		{
			memcpy(newEntry, oldEntry, sizeof(DistEntry *)*(slotNo));
		}

		if (slotNo < noOfEntry-1)
		{
			memcpy(&(newEntry[slotNo]), &(oldEntry[slotNo+1]), sizeof(DistEntry *)*(noOfEntry-slotNo-1));
		}

		distSlot->noOfEntry--;

		// update
		for (i = slotNo+1; i < noOfEntry; i++)
		{
			oldEntry[i]->distSlotNo--;
		}

		distSlot->entry = newEntry;

		free(oldEntry);
	}
}

void kDist_removeDistEntryFromDistSlot(DistEntry *distEntry)
{
	DistSlot *distSlot;
	int slotNo;

	distSlot = distEntry->distSlot;
	slotNo = distEntry->distSlotNo;

	kDist_removeEntryFromDistSlot(distSlot, slotNo);
}

// NULL: can't find
DistEntry *kDist_findDistEntry(Hash *hash, int id1, int id2)
{
	int newId1, newId2;
	int findId1, findId2;

	HashSlot *hashSlot;
	DistEntry **entry;
	DistEntry *distEntry;

	int slotNo;

	int i;

	if (id1 < id2)
	{
		newId1 = id1;
		newId2 = id2;
	}
	else
	{
		newId1 = id2;
		newId2 = id1;
	}

	slotNo = (id1+id2)%hash->noOfSlot;

	hashSlot = &(hash->slot[slotNo]);

	entry = hashSlot->entry;
	for (i = 0; i < hashSlot->noOfEntry; i++)
	{
		distEntry = entry[i];

		findId1 = distEntry->eqTuple1->id;
		findId2 = distEntry->eqTuple2->id;

		if ((findId1 == newId1) && (findId2 == newId2))
		{
			return distEntry;
		}

	}

	return NULL;
}

// -1: can't find
int kDist_findEqTupleIndex(int id, equivalenceClass *proEqClass)
{
	int i;
	equivalenceTuple **eqClass;

	eqClass = proEqClass->eqClass;

	for (i = 0; i < proEqClass->noOfEqClass; i++)
	{
		//if (id == eqClass[i].id)
		if (id == eqClass[i]->id)
		{
			return i;
		}
	}

	return -1;
}


// suppress the eq. tuple
void kDist_suppressEquivalenceTuple(equivalenceTuple *eqTuple, generalizedAllTable *genAllTable, int noOfDim)
{
	int i, j;
	int genBottomLevel;

	int attrNo;
	int noOfLevel;

	//noOfDim = ;

	generalizedLevelTable *genLevelTable;
	generalizedDimTable *genDimTable;

	for (i = 0; i < noOfDim-1; i++)
	{
		attrNo = i;		

		genDimTable = &(genAllTable->genDimTable[attrNo]);
		noOfLevel = genDimTable->noOfLevel;

		genBottomLevel = eqTuple->content[attrNo].level;
		
		for (j = genBottomLevel; j < noOfLevel-1;j++)
		{
			genLevelTable = &(genDimTable->gen[j]);

			generalizeEquivalenceTuple(eqTuple, attrNo, genLevelTable);
		}
	}
}



void kDist_removeAllUnsatisfiedDistEntryFromHashAndDistStorage(Hash *hash, DistStorage *distStorage, int id, equivalenceClass *finalEqClass, equivalenceClass *proEqClass, generalizedAllTable *genAllTable, int noOfDim, int *ind)
{
	int noOfFinalEqClass, noOfProEqClass;
	int i;

	int id2;

	int index;

	equivalenceTuple **finalEqTupleList;
	equivalenceTuple **proEqTupleList;

	DistEntry *distEntry;

	noOfFinalEqClass = finalEqClass->noOfEqClass;
	noOfProEqClass = proEqClass->noOfEqClass;

	finalEqTupleList = finalEqClass->eqClass;
	proEqTupleList = proEqClass->eqClass;


	// find the index of eqTuple with id
	index = kDist_findEqTupleIndex(id, proEqClass);

	(*ind) = index;

	// relationship within proEqClass
	for (i = 0; i < noOfProEqClass; i++)
	{
		//id2 = proEqTupleList[i].id;
		id2 = proEqTupleList[i]->id;
		if (id != id2)
		{
			// find the distEntry (id, id2) in hash
			distEntry = kDist_findDistEntry(hash, id, id2);

			if (distEntry != NULL)
			{
				// remove the distEntry (id, id2) from hash
				kDist_removeDistEntryFromHashSlot(distEntry);

				// remove the distEntry (id, id2) from distStorage
				kDist_removeDistEntryFromDistSlot(distEntry);

//printf("@@@@@@@@@@@@@@@@@@@@@ Removing (%d, %d) @@@@@@@@@@@@@@@@@@@@@\n", id, id2);
//fprintf(g_debugFilePtr, "@@@@@@@@@@@@@@@@@@@@@ Removing (%d, %d) @@@@@@@@@@@@@@@@@@@@@\n", id, id2);
//printDistStorage(g_debugFilePtr, distStorage);
			}
			else
			{
				//printf("Cannot find entry in the hash!\n");
			}

			// free the distEntry
			free(distEntry);
		}
	}

	// relationship between ProEqClass and finalEqClass
	for (i = 0; i < noOfFinalEqClass; i++)
	{
		//id2 = finalEqTupleList[i].id;
		id2 = finalEqTupleList[i]->id;

		// find the distEntry (id, id2) in hash
		distEntry = kDist_findDistEntry(hash, id, id2);

		if (distEntry != NULL)
		{
			// remove the distEntry (id, id2) from hash
			kDist_removeDistEntryFromHashSlot(distEntry);

			// remove the distEntry (id, id2) from distStorage
			kDist_removeDistEntryFromDistSlot(distEntry);

//printf("@@@@@@@@@@@@@@@@@@@@@ Removing (%d, %d) @@@@@@@@@@@@@@@@@@@@@\n", id, id2);
//fprintf(g_debugFilePtr, "@@@@@@@@@@@@@@@@@@@@@ Removing (%d, %d) @@@@@@@@@@@@@@@@@@@@@\n", id, id2);
//printDistStorage(g_debugFilePtr, distStorage);
		}
		else
		{
			//printf("Cannot find entry in the hash!\n");
		}

		// free the distEntry
		free(distEntry);
	}


}


void kDist_removeAllSatisfiedDistEntryFromHashAndDistStorage(Hash *hash, DistStorage *distStorage, int id, equivalenceClass *finalEqClass, equivalenceClass *proEqClass, generalizedAllTable *genAllTable, int noOfDim, int *ind)
{
	int noOfFinalEqClass, noOfProEqClass;
	int i;

	int id2;

	int index;

	equivalenceTuple **finalEqTupleList;
	equivalenceTuple **proEqTupleList;

	DistEntry *distEntry;

	noOfFinalEqClass = finalEqClass->noOfEqClass;
	noOfProEqClass = proEqClass->noOfEqClass;

	finalEqTupleList = finalEqClass->eqClass;
	proEqTupleList = proEqClass->eqClass;


	// find the index of eqTuple with id
	index = kDist_findEqTupleIndex(id, finalEqClass);

	(*ind) = index;

	// relationship within proEqClass
	for (i = 0; i < noOfProEqClass; i++)
	{
		//id2 = proEqTupleList[i].id;
		id2 = proEqTupleList[i]->id;
		//if (id != id2)
		{
			// find the distEntry (id, id2) in hash
			distEntry = kDist_findDistEntry(hash, id, id2);

			if (distEntry != NULL)
			{
				// remove the distEntry (id, id2) from hash
				kDist_removeDistEntryFromHashSlot(distEntry);

				// remove the distEntry (id, id2) from distStorage
				kDist_removeDistEntryFromDistSlot(distEntry);
			}
			else
			{
//printf(" ####################3 No Entry ########################33\n");
//fprintf(g_debugFilePtr, " ####################3 No Entry ########################33\n");
//printDistStorage(g_debugFilePtr, distStorage);

				//printf("Cannot find entry in the hash!\n");
			}

			// free the distEntry
			free(distEntry);
		}
	}

}

void kDist_createAllDistEntry(equivalenceTuple **eqTuple, equivalenceClass *eqClass, Hash *hash, DistStorage *distStorage, generalizedAllTable *genAllTable, configure *conf)
{
	int i;
	int id2; 
	float dist;

	equivalenceTuple **entry;
	equivalenceTuple *eqTuple1;
	equivalenceTuple *eqTuple2;

	DistEntry *distEntry;

	int id;

	eqTuple1 = (*eqTuple);

	id = eqTuple1->id;
	
	entry = eqClass->eqClass;

	for (i = 0; i < eqClass->noOfEqClass; i++)
	{ 
		//id2 = entry[i].id;
		id2 = entry[i]->id;
		if (id2 != id)
		{
			// calculate the distance
			//eqTuple2 = &(entry[i]);
			eqTuple2 = entry[i];
			dist = calculateDistanceBetweenEquivalenceTuple(conf, genAllTable, eqTuple1, eqTuple2);
			
			// create the distEntry
			distEntry = distEntry_new(dist, &eqTuple1, &eqTuple2);

			// insert into hash and distStorage
			kDist_insertDistEntry2Hash(hash, &distEntry);
			kDist_insertDistEntry2DistStorage(distStorage, &distEntry);

		}
	}
}


// assume the last element (due to just inserted)
equivalenceTuple *kDist_getEqTupleFromProcFinal(equivalenceTuple *eqTuple, equivalenceClass *proEqClass, equivalenceClass *finalEqClass, int K)
{
	equivalenceTuple *returnValue;

	equivalenceClass *eqClass;

	if (eqTuple->noOfMember < K)
	{
		eqClass = proEqClass;
	}
	else
	{
		eqClass = finalEqClass;
	}

	//returnValue = &(eqClass->eqClass[eqClass->noOfEqClass-1]);
	returnValue = eqClass->eqClass[eqClass->noOfEqClass-1];

	return returnValue;
}

void kDist_createAllPairsDistEntry(equivalenceTuple **eqTuple, equivalenceClass *proEqClass, equivalenceClass *finalEqClass, Hash *hash, DistStorage *distStorage, generalizedAllTable *genAllTable, configure *conf)
{	
	int K, noOfDim;

	equivalenceTuple *persistEqTuple;

	K = conf->K;
	noOfDim = conf->noOfDim;

	//persistEqTuple = kDist_getEqTupleFromProcFinal(eqTuple, proEqClass, finalEqClass, K);	
	persistEqTuple = (*eqTuple);

	//if (eqTuple->noOfMember < K)
	if (persistEqTuple->noOfMember < K)
	{
		// relationship with proEqClass
		kDist_createAllDistEntry(&persistEqTuple, proEqClass, hash, distStorage, genAllTable, conf);

		// relationship with finalEqClass
		kDist_createAllDistEntry(&persistEqTuple, finalEqClass, hash, distStorage, genAllTable, conf);
	}
	else
	{
		// relationship with proEqClass
		kDist_createAllDistEntry(&persistEqTuple, proEqClass, hash, distStorage, genAllTable, conf);

		// no need for relationship with finalEqClass
	}


}


void kDist_createAllPairsDistEntrySuppression(equivalenceTuple **eqTuple, equivalenceClass *proEqClass, equivalenceClass *finalEqClass, Hash *hash, DistStorage *distStorage, generalizedAllTable *genAllTable, configure *conf)
{	
	int noOfDim, K;

	equivalenceTuple *persistEqTuple;

	noOfDim = conf->noOfDim;
	K = conf->K;

	//persistEqTuple = kDist_getEqTupleFromProcFinal(eqTuple, proEqClass, finalEqClass, K);	
	persistEqTuple = (*eqTuple);

	//if (eqTuple->noOfMember < K)
	if (persistEqTuple->noOfMember < K)
	{
		// relationship with proEqClass
		kDist_createAllDistEntry(&persistEqTuple, proEqClass, hash, distStorage, genAllTable, conf);

		// relationship with finalEqClass
		//kDist_createAllDistEntry(&persistEqTuple, finalEqClass, hash, distStorage, genAllTable, noOfDim);
	}
	else
	{
		// relationship with proEqClass
		kDist_createAllDistEntry(&persistEqTuple, proEqClass, hash, distStorage, genAllTable, conf);

		// no need for relationship with finalEqClass
	}
}

void kDist_removeAllAnyDistEntryFromHashAndDistStorage(int noOfMember, int K, Hash *hash, DistStorage *distStorage, int id, equivalenceClass *finalEqClass, equivalenceClass *proEqClass, generalizedAllTable *genAllTable, int noOfDim, int *ind)
{
	if (noOfMember < K)
	{	
		kDist_removeAllUnsatisfiedDistEntryFromHashAndDistStorage(hash, distStorage, id, finalEqClass, proEqClass, genAllTable, noOfDim, ind);
	}
	else
	{
		kDist_removeAllSatisfiedDistEntryFromHashAndDistStorage(hash, distStorage, id, finalEqClass, proEqClass, genAllTable, noOfDim, ind);
	}
}


float kDist_findMinDist_proEqTuple(int id, Hash *hash, equivalenceClass *proEqClass, equivalenceClass *finalEqClass, int K)
{
	int i;

	int id2;

	int noOfProEqClass, noOfFinalEqClass;

	equivalenceTuple **proEqTuple;
	equivalenceTuple **finalEqTuple;

	DistEntry *distEntry;
	float dist, minDist;

	int isStart = 0;

	minDist = -1;

	noOfProEqClass = proEqClass->noOfEqClass;
	noOfFinalEqClass = finalEqClass->noOfEqClass;

	proEqTuple = proEqClass->eqClass;
	finalEqTuple = finalEqClass->eqClass;

	// relationship with proEqClass
	for (i = 0; i < noOfProEqClass; i++)
	{
		//id2 = proEqTuple[i].id;
		id2 = proEqTuple[i]->id;

		if (id2 != id)
		{
			distEntry = kDist_findDistEntry(hash, id, id2);
			if (distEntry != NULL)
			{
				dist = distEntry->dist;

				if (isStart == 0)
				{
					minDist = dist;
					isStart = 1;
				}

				if (dist < minDist)
				{
					minDist = dist;
				}
			}
			else
			{
				// do nothing
			}
		}
	}

	// relationship with finalEqclass
	for (i = 0; i < noOfFinalEqClass; i++)
	{
		//id2 = finalEqTuple[i].id;
		id2 = finalEqTuple[i]->id;

		if (id2 != id)
		{
			distEntry = kDist_findDistEntry(hash, id, id2);
			if (distEntry != NULL)
			{
				dist = distEntry->dist;

				if (isStart == 0)
				{
					minDist = dist;
					isStart = 1;
				}

				if (dist < minDist)
				{
					minDist = dist;
				}
			}
			else
			{
				// do nothing
			}
		}
	}

	return minDist;
}


void kDist_suppressBatch(Hash *hash, DistStorage *distStorage, int id, equivalenceClass *finalEqClass, equivalenceClass *proEqClass, generalizedAllTable *genAllTable, configure *conf)
{
	int index;

	int noOfDim, K;

	equivalenceTuple eqTuple;
	equivalenceTuple *realEqTuple;

	int noOfMember;

	noOfDim = conf->noOfDim;
	K = conf->K;

	// need to suppress the equiv. class
	kDist_removeAllUnsatisfiedDistEntryFromHashAndDistStorage(hash, distStorage, id, finalEqClass, proEqClass, genAllTable, noOfDim, &index);

	// suppress the equivalence tuple
	memcpy(&eqTuple, proEqClass->eqClass[index], sizeof(equivalenceTuple));
	kDist_suppressEquivalenceTuple(&eqTuple, genAllTable, noOfDim);

	// move the eq. tuple from proEqClass to finalEqClass
	removeEqTupleFromEqClass(proEqClass, index);
	insertEqTupleIntoEqClass(finalEqClass, &eqTuple);


	// remove eqtuple
	noOfMember = eqTuple.noOfMember;
	id = eqTuple.id;
	kDist_removeAllAnyDistEntryFromHashAndDistStorage(noOfMember, conf->K, hash, distStorage, id, finalEqClass, proEqClass, genAllTable, noOfDim, &index);

//	fprintf(g_debugFilePtr, "  ~~~~~~~~~~~~~~~~~~~ Before Creating.....\n");
//	printf("  ~~~~~~~~~~~~~~~~~~~ Before Creating.....\n");
//	printDistStorage(g_debugFilePtr, distStorage);

	// no need to insert distEntry between this new eq. tuple and other tuples
	// because other tuples will be suppressed automatically (if there is a need)
	realEqTuple = finalEqClass->eqClass[finalEqClass->noOfEqClass-1];
	kDist_createAllPairsDistEntrySuppression(&realEqTuple, proEqClass, finalEqClass, hash, distStorage, genAllTable, conf);

//	fprintf(g_debugFilePtr, "  ~~~~~~~~~~~~~~~~~~~ After Creating.....\n");
//	printf("  ~~~~~~~~~~~~~~~~~~~ After Creating.....\n");
//	printDistStorage(g_debugFilePtr, distStorage);
}


void kDist_suppressBatchRandom(int id, equivalenceClass *finalEqClass, equivalenceClass *proEqClass, generalizedAllTable *genAllTable, configure *conf)
{
	int index;

	int noOfDim, K;

	equivalenceTuple eqTuple;
	equivalenceTuple *realEqTuple;

	int noOfMember;

	noOfDim = conf->noOfDim;
	K = conf->K;

	// need to suppress the equiv. class
	//kDist_removeAllUnsatisfiedDistEntryFromHashAndDistStorage(hash, distStorage, id, finalEqClass, proEqClass, genAllTable, noOfDim, &index);

	// the last elt (which has only ONE in proEqClass)
	index = 0;

	// suppress the equivalence tuple
	memcpy(&eqTuple, proEqClass->eqClass[index], sizeof(equivalenceTuple));
	kDist_suppressEquivalenceTuple(&eqTuple, genAllTable, noOfDim);

	// move the eq. tuple from proEqClass to finalEqClass
	removeEqTupleFromEqClass(proEqClass, index);
	insertEqTupleIntoEqClass(finalEqClass, &eqTuple);


	// remove eqtuple
	noOfMember = eqTuple.noOfMember;
	id = eqTuple.id;
	//kDist_removeAllAnyDistEntryFromHashAndDistStorage(noOfMember, conf->K, hash, distStorage, id, finalEqClass, proEqClass, genAllTable, noOfDim, &index);

//	fprintf(g_debugFilePtr, "  ~~~~~~~~~~~~~~~~~~~ Before Creating.....\n");
//	printf("  ~~~~~~~~~~~~~~~~~~~ Before Creating.....\n");
//	printDistStorage(g_debugFilePtr, distStorage);

	// no need to insert distEntry between this new eq. tuple and other tuples
	// because other tuples will be suppressed automatically (if there is a need)
	realEqTuple = finalEqClass->eqClass[finalEqClass->noOfEqClass-1];
	//kDist_createAllPairsDistEntrySuppression(&realEqTuple, proEqClass, finalEqClass, hash, distStorage, genAllTable, conf);

//	fprintf(g_debugFilePtr, "  ~~~~~~~~~~~~~~~~~~~ After Creating.....\n");
//	printf("  ~~~~~~~~~~~~~~~~~~~ After Creating.....\n");
//	printDistStorage(g_debugFilePtr, distStorage);
}



void kDist_calculatePairDistanceWithSuppression(Hash *hash, DistStorage *distStorage, equivalenceClass *finalEqClass, equivalenceClass *proEqClass, generalizedAllTable *genAllTable, configure *conf)
{
	int noOfFinalEqClass, noOfProEqClass;
	int noOfDim;
	int i, j;

	int id1, id2;

	float dist;
	float minDist;

	int id;
	int K;

	float threshold;

	DistEntry *distEntry;

	equivalenceTuple *eqTuple1;
	equivalenceTuple *eqTuple2;

	equivalenceTuple **finalEqTupleList;
	equivalenceTuple **proEqTupleList;

	int *isSuppress;

	noOfDim = conf->noOfDim;
	K = conf->K;

	noOfFinalEqClass = finalEqClass->noOfEqClass;
	noOfProEqClass = proEqClass->noOfEqClass;

	finalEqTupleList = finalEqClass->eqClass;
	proEqTupleList = proEqClass->eqClass;

	isSuppress = (int *) malloc(sizeof(int)*noOfProEqClass);
	memset(isSuppress, 0, sizeof(int)*noOfProEqClass);

	for (i = 0; i < noOfProEqClass; i++)
	{
		if (i%100 == 0)
		{
			printf("   ~~~ %d ~~~~\n", i);
		}
		minDist = -1;

		//eqTuple1 = &(proEqTupleList[i]);
		eqTuple1 = proEqTupleList[i];

		id1 = eqTuple1->id;

		// also relationship within ProEqClass
		// find the previous records in order to update minDist
		minDist = kDist_findMinDist_proEqTuple(id1, hash, proEqClass, finalEqClass, K);


		// relationship within ProEqClass
		for (j = i+1; j < noOfProEqClass; j++)
		{
			//eqTuple2 = &(proEqTupleList[j]);
			eqTuple2 = proEqTupleList[j];
			
			id2 = eqTuple2->id;

			if (id1 != id2)
			{
				// calculate the distance
				
				dist = calculateDistanceBetweenEquivalenceTuple(conf, genAllTable, eqTuple1, eqTuple2);
				if ((minDist == -1) || (dist < minDist))
				{
					minDist = dist;
				}

				// create the distEntry
				distEntry = distEntry_new(dist, &eqTuple1, &eqTuple2);

				// insert into hash and DistStorage
				kDist_insertDistEntry2Hash(hash, &distEntry);
				kDist_insertDistEntry2DistStorage(distStorage, &distEntry);

				//fprintf(g_debugFilePtr, "  ~~~~~~~~~~~~~~~~~~~ Inserting (%d, %d).....\n", id1, id2);
				//printf("  ~~~~~~~~~~~~~~~~~~~ Inserting (%d, %d).....\n", id1, id2);
				//printDistStorage(g_debugFilePtr, distStorage);


			}
		}

		


		// relationship between ProEqClass and finalEqClass
		for (j = 0; j < noOfFinalEqClass; j++)
		{
			//eqTuple2 = &(finalEqTupleList[j]);
			eqTuple2 = finalEqTupleList[j];

			id2 = eqTuple2->id;

			// calculate the distance
			
			dist = calculateDistanceBetweenEquivalenceTuple(conf, genAllTable, eqTuple1, eqTuple2);
			if ((minDist == -1) || (dist < minDist))
			{
				minDist = dist;
			}

			// create the distEntry
			distEntry = distEntry_new(dist, &eqTuple1, &eqTuple2);

			// insert into hash and DistStorage
			kDist_insertDistEntry2Hash(hash, &distEntry);
			kDist_insertDistEntry2DistStorage(distStorage, &distEntry);

			//fprintf(g_debugFilePtr, "  ~~~~~~~~~~~~~~~~~~~ Inserting (%d, %d).....\n", id1, id2);
			//printf("  ~~~~~~~~~~~~~~~~~~~ Inserting (%d, %d).....\n", id1, id2);
			//printDistStorage(g_debugFilePtr, distStorage);
		}

		//**** check the suppression

		// minus the sensitive attr. 
		
		threshold = (float) ((noOfDim-1)*eqTuple1->noOfMember);
		if (minDist > threshold)
		{
			// need to suppress the equivalence class
			isSuppress[i] = 1;
		}
	}

	// suppress the equivalence class
	for (i = noOfProEqClass-1; i >= 0; i--)
	{
		if (isSuppress[i] == 1)
		{
			

//fprintf(g_debugFilePtr, "  ~~~~~~~~~~~~~~~~~~~ Before Suppress.....\n");
//printf("  ~~~~~~~~~~~~~~~~~~~ Before Suppress.....\n");
//printDistStorage(g_debugFilePtr, distStorage);

			id = eqTuple1->id;

//fprintf(g_debugFilePtr, "   &&&&&&&&&&&&&&&&&& Suppressing (%d)..... &&&&&&&&&&&&&&&&&&\n", id);
//printf("   &&&&&&&&&&&&&&&&&& Suppressing (%d)..... &&&&&&&&&&&&&&&&&&\n", id);

			kDist_suppressBatch(hash, distStorage, id, finalEqClass, proEqClass, genAllTable, conf);

			//fprintf(g_debugFilePtr, "  ~~~~~~~~~~~~~~~~~~~ After Suppress.....\n");
			//printf("  ~~~~~~~~~~~~~~~~~~~ After Suppress.....\n");
			//printDistStorage(g_debugFilePtr, distStorage);
		}
	}

	free(isSuppress);
}


AttributeList *kDist_getAttributeList(int noOfDim, equivalenceTuple *eqTuple1, equivalenceTuple *eqTuple2)
{
	AttributeList *returnValue;

	int noOfDiff;
	int *attrList;

	dimension *dim1;
	dimension *dim2;


	int i;

	returnValue = (AttributeList *) malloc(sizeof(AttributeList));
	memset(returnValue, 0, sizeof(AttributeList));


	dim1 = eqTuple1->content;
	dim2 = eqTuple2->content;

	// for each dim (minus the last sensitive attr.)

	// count the no. of attr different
	noOfDiff = 0;
	for (i = 0; i < noOfDim-1; i++)
	{
		if (dim1[i].level != dim2[i].level)
		{
			noOfDiff++;
		}
		else
		{
			if (dim1[i].value != dim2[i].value)
			{
				noOfDiff++;
			}
		}
	}

	// create the list
	returnValue->noOfAttr = noOfDiff;
	attrList = (int *) malloc(sizeof(int)*noOfDiff);

	// insert the elemt to the list
	noOfDiff = 0;
	for (i = 0; i < noOfDim-1; i++)
	{
		if (dim1[i].level != dim2[i].level)
		{
			attrList[noOfDiff] = i;
			noOfDiff++;
		}
		else
		{
			if (dim1[i].value != dim2[i].value)
			{
				attrList[noOfDiff] = i;
				noOfDiff++;
			}
		}
	}	

	returnValue->attr = attrList;

	return returnValue;
}


// NULL: nothing can be found
DistEntry *kDist_findDistEntryWithMinDist(DistStorage *distStorage, generalizedAllTable *genAllTable, int noOfDim)
{
	int i, j;
	int attrNo;
	int slotNo;

	DistSlot *distSlot;
	DistEntry **distEntry;

	DistEntry *aDistEntry;

	equivalenceTuple *eqTuple1;
	equivalenceTuple *eqTuple2;

	float minDist, dist;

	int noOfMin;

	int index;
	int *smallIndex;

	int *minIndex;

	int noOfEntry;

	AttributeList *attrList;
	AttributeList *newAttrList;

	int minNoOfAttr, noOfAttr;

	int noOfSmall;

	int smallestIndex;

	int indexMin;

	int eval, minEval;
	int noOfPoss;

	slotNo = -1;


	for (i = 0; i < distStorage->noOfSlot; i++)
	{
		distSlot = &(distStorage->slot[i]);
		if (distSlot->noOfEntry != 0)
		{
			slotNo = i;
			break;
		}
	}

	if (slotNo != -1)
	{
		// found

		noOfEntry = distSlot->noOfEntry;
		distEntry = distSlot->entry;

		minIndex = (int *) malloc(sizeof(int)*noOfEntry);
		memset(minIndex, 0, sizeof(int)*noOfEntry);
		noOfMin = 0;

		// assign the first (because it should contain at least one elt)
		minDist = distEntry[0]->dist;

		// find the min. dist
		for (i = 1; i < noOfEntry; i++)
		{
			dist = distEntry[i]->dist;
			if (dist < minDist)
			{
				minDist = dist;
			}
		}

		// locate which entries has the min. dist
		for (i = 0; i < noOfEntry; i++)
		{
			dist = distEntry[i]->dist;
			if (dist == minDist)
			{
				minIndex[noOfMin] = i;
				noOfMin++;
			}
		}

		// calculate the attribute list
		attrList = (AttributeList *) malloc(sizeof(AttributeList)*noOfMin);
		memset(attrList, 0, sizeof(AttributeList)*noOfMin);
		for (i = 0; i < noOfMin; i++)
		{
			index = minIndex[i];

			eqTuple1 = distEntry[index]->eqTuple1;
			eqTuple2 = distEntry[index]->eqTuple2;

			newAttrList = kDist_getAttributeList(noOfDim, eqTuple1, eqTuple2);

			memcpy(&(attrList[i]), newAttrList, sizeof(AttributeList));

			// free the newAttr. list
			free(newAttrList);
		}

		// among a number of the attribute list
		// find the smallest size
		minNoOfAttr = attrList[0].noOfAttr;		// as there is at least one elt
		for (i = 1; i < noOfMin; i++)
		{
			noOfAttr = attrList[i].noOfAttr ;
			if (noOfAttr < minNoOfAttr)
			{
				minNoOfAttr = noOfAttr;
			}
		}

		// get the list of elts with the smallest size
		smallIndex = (int *) malloc(sizeof(int)*noOfMin);
		noOfSmall = 0;
		for (i = 0; i < noOfMin; i++)
		{
			noOfAttr = attrList[i].noOfAttr ;
			if (noOfAttr == minNoOfAttr)
			{
				smallIndex[noOfSmall] = i;
				noOfSmall++;
			}
		}

		// among the attr. list (with smallest size)
		// choose the attr. with greatest poss value
		for (i = 0; i < noOfSmall; i++)
		{
			indexMin = smallIndex[i];
			
			for (j = 0; j < attrList[indexMin].noOfAttr; j++)
			{
				// find the attr. no
				attrNo = attrList[indexMin].attr[j];

				// 
				noOfPoss = genAllTable->genDimTable[attrNo].gen[0].noOfPoss;
				eval += noOfPoss;
			}

			if (i == 0)
			{
				minEval = eval;
				smallestIndex = i;
			}

			if (eval < minEval)
			{
				minEval = eval;
				smallestIndex = i;
			}
		}

		indexMin = smallIndex[smallestIndex];
		index = minIndex[indexMin];

		// get the dist entry
		aDistEntry = distEntry[index];



		free(smallIndex);


		// free the attr list
		for (i = 0; i < noOfMin; i++)
		{
			free(attrList[i].attr);
		}
		free(attrList);

		free(minIndex);
	}
	else
	{
		// nothing stored in DistStorage
		aDistEntry = NULL;
	}

	return aDistEntry;
}


// return the new eq. tuple with newNoOfMember
// NULL: when newNoOfMember >= noOfMember
equivalenceTuple *kDist_splitEquivalenceTuple(equivalenceTuple *eqTuple, int newNoOfMember, int noOfDim)
{
	equivalenceTuple *returnValue;
	int noOfMember;
	int newNoOfSen;

	int *changeMember;
	int *changeIsSen;

	int i;


	noOfMember = eqTuple->noOfMember;

	if (newNoOfMember < noOfMember )
	{
		// for the new eq. tuple
		returnValue = (equivalenceTuple *) malloc(sizeof(equivalenceTuple));

		returnValue->content = (dimension *) malloc(sizeof(dimension)*noOfDim);
		memcpy(returnValue->content, eqTuple->content, sizeof(dimension)*noOfDim);
		returnValue->noOfMember = newNoOfMember;
		returnValue->id = eqTuple->id;
		returnValue->member = (int *) malloc(sizeof(int)*newNoOfMember);
		returnValue->isSen = (int *) malloc(sizeof(int)*newNoOfMember);
		memcpy(returnValue->member, eqTuple->member, sizeof(int)*newNoOfMember);
		memcpy(returnValue->isSen, eqTuple->isSen, sizeof(int)*newNoOfMember);
		

		newNoOfSen = 0;
		for (i = 0; i < newNoOfMember; i++)
		{
			if (eqTuple->isSen[i] == 1)
			{
				newNoOfSen++;
			}
		}
		returnValue->noOfSen = newNoOfSen;

		// for the original eq. tuple
		// minus the original tuple

		changeMember = (int *) malloc(sizeof(int)*(noOfMember-newNoOfMember));
		changeIsSen = (int *) malloc(sizeof(int)*(noOfMember-newNoOfMember));
		memcpy(changeMember, &(eqTuple->member[newNoOfMember]), sizeof(int)*(noOfMember-newNoOfMember));
		memcpy(changeIsSen, &(eqTuple->isSen[newNoOfMember]), sizeof(int)*(noOfMember-newNoOfMember));

		eqTuple->noOfMember = noOfMember-newNoOfMember;
		eqTuple->noOfSen = eqTuple->noOfSen-newNoOfSen;

		// free
		free(eqTuple->member);
		free(eqTuple->isSen);

		eqTuple->member = changeMember;
		eqTuple->isSen = changeIsSen;
		
		return returnValue;
	}
	else
	{
		return NULL;
	}
}

void kDist_removeEqTupleFromProcFinal(equivalenceClass *proEqClass, equivalenceClass *finalEqClass, equivalenceTuple *eqTuple, int K)
{
	int id;
	int index;

	id = eqTuple->id;
	if (eqTuple->noOfMember < K)
	{
		index = kDist_findEqTupleIndex(id, proEqClass);

		// handle the case for suppression
		if (index == -1)
		{
			index = kDist_findEqTupleIndex(id, finalEqClass);
			removeEqTupleFromEqClass(finalEqClass, index);
		}
		else
		{
			removeEqTupleFromEqClass(proEqClass, index);
		}
	}
	else
	{
		index = kDist_findEqTupleIndex(id, finalEqClass);
		removeEqTupleFromEqClass(finalEqClass, index);
	}
}

void kDist_insertEqTuple2ProcFinal(equivalenceClass *proEqClass, equivalenceClass *finalEqClass, equivalenceTuple *eqTuple, int K)
{

	if (eqTuple->noOfMember < K)
	{
		insertEqTupleIntoEqClass(proEqClass, eqTuple);
	}
	else
	{
		insertEqTupleIntoEqClass(finalEqClass, eqTuple);
	}
}






void free_eqClass(equivalenceTuple *eqTuple)
{
	free(eqTuple->member);
	free(eqTuple->isSen);
	free(eqTuple->content);
}

void kDist_findMinDistForGeneralization(DistStorage *distStorage, Hash *hash, generalizedAllTable *genAllTable, configure *conf, equivalenceClass *proEqClass, equivalenceClass *finalEqClass)
{
	DistEntry *bestDistEntry;

	equivalenceTuple *eqTuple1;
	equivalenceTuple *eqTuple2;

	equivalenceTuple *newEqTuple;

	int K;

	int noOfDim;

	int id1, id2;
	int index;
	int noOfMember1, noOfMember2;

	int stubNoOfMember;
	equivalenceTuple *stubEqTuple;

	equivalenceTuple *tempEqTuple;
	int tempNo;

	noOfDim = conf->noOfDim;
	K = conf->K;

	bestDistEntry = kDist_findDistEntryWithMinDist(distStorage, genAllTable, noOfDim);

	eqTuple1 = bestDistEntry->eqTuple1;
	eqTuple2 = bestDistEntry->eqTuple2;

	id1 = eqTuple1->id;
	id2 = eqTuple2->id;

printf("  ************************ Merging (%d, %d) **********************\n", id1, id2);
fprintf(g_debugFilePtr, "  ************************ Merging (%d, %d) **********************\n", id1, id2);

//fprintf(g_debugFilePtr, "        ************** Before Removal ************** (After Merging (%d, %d)....)\n", eqTuple1->id, eqTuple2->id);
//fprintf(g_debugFilePtr, "Printing Dist Storage...\n");
//printf("        ************** Before Removal ************** (After Merging (%d, %d)....)\n", eqTuple1->id, eqTuple2->id);
//printf("Printing Dist Storage...\n");
//printDistStorage(g_debugFilePtr, distStorage);

	noOfMember1 = eqTuple1->noOfMember;
	noOfMember2 = eqTuple2->noOfMember;


	// remove eqTuple1 from hash/diststorage
	kDist_removeAllAnyDistEntryFromHashAndDistStorage(noOfMember1, conf->K, hash, distStorage, id1, finalEqClass, proEqClass, genAllTable, noOfDim, &index);

	// remove eqTuple2 from hash/diststorage
	kDist_removeAllAnyDistEntryFromHashAndDistStorage(noOfMember2, conf->K, hash, distStorage, id2, finalEqClass, proEqClass, genAllTable, noOfDim, &index);



	if (noOfMember1+noOfMember2 < 2*conf->K)
	{
		// merge two equivalence classes
		newEqTuple = mergeTwoEquivalenceTuple(noOfDim, genAllTable, eqTuple1, eqTuple2);

		// remove eq. tuple 1
		kDist_removeEqTupleFromProcFinal(proEqClass, finalEqClass, eqTuple1, K);

		// remove eq. tuple 2
		kDist_removeEqTupleFromProcFinal(proEqClass, finalEqClass, eqTuple2, K);

//fprintf(g_debugFilePtr, "        ************** After Removal ************** (After Merging (%d, %d)....)\n", eqTuple1->id, eqTuple2->id);
//fprintf(g_debugFilePtr, "Printing Dist Storage...\n");
//printf("        ************** After Removal ************** (After Merging (%d, %d)....)\n", eqTuple1->id, eqTuple2->id);
//printf("Printing Dist Storage...\n");
//printDistStorage(g_debugFilePtr, distStorage);

		// insert newEqTuple to proEqClass/finalEqClass
		kDist_insertEqTuple2ProcFinal(proEqClass, finalEqClass, newEqTuple, K);

		// generate all pairs distEntry (from newEqtuple)
		kDist_createAllPairsDistEntry(&newEqTuple, proEqClass, finalEqClass, hash, distStorage, genAllTable, conf);

//fprintf(g_debugFilePtr, "        ************** After Insertion ************** (After Merging (%d, %d)....)\n", eqTuple1->id, eqTuple2->id);
//fprintf(g_debugFilePtr, "Printing Dist Storage...\n");
//printf("        ************** After Insertion ************** (After Merging (%d, %d)....)\n", eqTuple1->id, eqTuple2->id);
//printf("Printing Dist Storage...\n");
//printDistStorage(g_debugFilePtr, distStorage);

		// free
		//free(newEqTuple);

		// 
		free_eqClass(eqTuple1);
		free_eqClass(eqTuple2);
		free(eqTuple1);
		free(eqTuple2);
	}
	else
	{
		if (noOfMember1 < K)
		{
			// do nothing
		}
		else
		{
			// swap
			tempEqTuple = eqTuple1;
			eqTuple1 = eqTuple2;
			eqTuple2 = tempEqTuple;

			tempNo = noOfMember1;
			noOfMember1 = noOfMember2;
			noOfMember2 = tempNo;
		}

			// split eqTuple2
			stubNoOfMember = K-noOfMember1;
			stubEqTuple = kDist_splitEquivalenceTuple(eqTuple2, stubNoOfMember, noOfDim);

			// merge stubEqTuple to eqTuple1
			newEqTuple = mergeTwoEquivalenceTuple(noOfDim, genAllTable, eqTuple1, stubEqTuple);

			// remove eq. tuple 1
			kDist_removeEqTupleFromProcFinal(proEqClass, finalEqClass, eqTuple1, K);

			// remove eq. tuple 2
			kDist_removeEqTupleFromProcFinal(proEqClass, finalEqClass, eqTuple2, K);

			// insert newEqTuple to proEqClass/finalEqClass
			kDist_insertEqTuple2ProcFinal(proEqClass, finalEqClass, newEqTuple, K);

			// insert eqTuple2 to proEqClass/finalEqClass
			kDist_insertEqTuple2ProcFinal(proEqClass, finalEqClass, eqTuple2, K);

			// generate all pairs (from newEqTuple)
			kDist_createAllPairsDistEntry(&newEqTuple, proEqClass, finalEqClass, hash, distStorage, genAllTable, conf);

			// generate all pairs (from eqTuple2)
			kDist_createAllPairsDistEntry(&eqTuple2, proEqClass, finalEqClass, hash, distStorage, genAllTable, conf);

			// free
			//free(newEqTuple);
			//free(stubEqTuple);

			// 
			free_eqClass(eqTuple1);
			free_eqClass(eqTuple2);
			free(eqTuple1);
			free(eqTuple2);	
	}
}


void kDist_moveSatisfiedEqClass(equivalenceClass *proEqClass, equivalenceClass *finalEqClass, int K)
{
	int noOfProEqClass, noOfFinalEqClass;
	equivalenceTuple **proEqTuple;
	equivalenceTuple **finalEqTuple;

	equivalenceTuple *aEqTuple;

	int *removeList;
	int noOfRemove;

	int index;

	int i;

	noOfProEqClass = proEqClass->noOfEqClass;
	proEqTuple = proEqClass->eqClass;

	noOfFinalEqClass = finalEqClass->noOfEqClass;
	finalEqTuple = finalEqClass->eqClass;

	removeList = (int *) malloc(sizeof(int)*noOfProEqClass);

	noOfRemove = 0;
	for (i = 0; i < noOfProEqClass; i++)
	{
		//if (proEqTuple[i].noOfMember >= K)
		if (proEqTuple[i]->noOfMember >= K)
		{
			//aEqTuple = &(proEqTuple[i]);
			aEqTuple = proEqTuple[i];

			// insert to finalEqClass
			insertEqTupleIntoEqClass(finalEqClass, aEqTuple);

			// remove from proEqClass
			removeList[noOfRemove] = i;
			noOfRemove++;
		}
	}

	for (i = noOfRemove-1; i >= 0; i--)
	{
		index = removeList[i];
		removeEqTupleFromEqClass(proEqClass, index);
	}
	free(removeList);
}



void kDist_suppressIteration(Hash *hash, DistStorage *distStorage, equivalenceClass *proEqClass, equivalenceClass *finalEqClass, configure *conf, generalizedAllTable *genAllTable)
{
	int id;
	int i;

	equivalenceTuple **proEqTuple;
	equivalenceTuple *eqTupleOne;

	float minDist;

	int noOfDim;
	int K;

	float threshold;

	noOfDim = conf->noOfDim;
	K = conf->K;

	proEqTuple = proEqClass->eqClass;


	//float kDist_findMinDist_proEqTuple(int id, equivalenceClass *proEqClass, equivalenceClass *finalEqClass, int K)

	for (i = 0; i < proEqClass->noOfEqClass; i++)
	{
		//eqTupleOne = &(proEqTuple[i]);
		eqTupleOne = proEqTuple[i];
		id = eqTupleOne->id;

		// find the min. dist
		minDist = kDist_findMinDist_proEqTuple(id, hash, proEqClass, finalEqClass, K);

		// check suppression
		threshold = (float) ((noOfDim-1)*eqTupleOne->noOfMember);

		if (minDist > threshold)
		{
			kDist_suppressBatch(hash, distStorage, id, finalEqClass, proEqClass, genAllTable, conf);
			proEqTuple = proEqClass->eqClass;
		}
	}
}


float *kDist_getPairDistance(int targetIndex, equivalenceClass *finalEqClass, equivalenceClass *proEqClass, generalizedAllTable *genAllTable, configure *conf)
{
	float *returnValue;

	int j;

	int id1, id2;

	int noOfFinalEqClass, noOfProEqClass;

	equivalenceTuple **finalEqTuple;
	equivalenceTuple **proEqTuple;

	equivalenceTuple *eqTuple1;
	equivalenceTuple *eqTuple2;

	float dist;

	int total;

	int index;


	noOfFinalEqClass = finalEqClass->noOfEqClass;
	noOfProEqClass = proEqClass->noOfEqClass;

	total = noOfProEqClass+noOfFinalEqClass;

	finalEqTuple = finalEqClass->eqClass;
	proEqTuple = proEqClass->eqClass;	

	returnValue = (float *) malloc(sizeof(float)*total);
	memset(returnValue, 0, sizeof(float)*total);

	eqTuple1 = proEqTuple[targetIndex];
	id1 = eqTuple1->id;

	index = 0;
	for (j = 0; j < noOfProEqClass; j++)
	{
		eqTuple2 = proEqTuple[j];
		id2 = eqTuple2->id;
		if (id1 != id2)
		{
			dist = calculateDistanceBetweenEquivalenceTuple(conf, genAllTable, eqTuple1, eqTuple2);
			returnValue[index] = dist;
		}

		index++;
	}
	
	for (j = 0; j < noOfFinalEqClass; j++)
	{
		eqTuple2 = finalEqTuple[j];
		id2 = eqTuple2->id;
		if (id1 != id2)
		{
			dist = calculateDistanceBetweenEquivalenceTuple(conf, genAllTable, eqTuple1, eqTuple2);
			returnValue[index] = dist;
		}

		index++;
	}

	return returnValue;
}

int kDist_getMinDistIndex(int targetIndex, equivalenceClass *proEqClass, equivalenceClass *finalEqClass, float *dist, generalizedAllTable *genAllTable, int noOfDim)
{
	int *minIndex;
	int noOfMin;
	int i, j;

	float minDist;

	int index;

	int tempIndex;

	equivalenceTuple **proEqTuple;
	equivalenceTuple **finalEqTuple;
	equivalenceTuple *eqTuple1;
	equivalenceTuple *eqTuple2;

	AttributeList *attrList;
	AttributeList *newAttrList;

	int noOfProEqTuple, noOfFinalEqTuple;

	int minNoOfAttr, noOfAttr;

	int *smallIndex;

	int noOfSmall;
	int smallestIndex;

	int indexMin;

	int attrNo;

	int total;

	int eval, minEval;
	int noOfPoss;

	noOfProEqTuple = proEqClass->noOfEqClass;
	noOfFinalEqTuple = finalEqClass->noOfEqClass;

	total = noOfProEqTuple+noOfFinalEqTuple;

	proEqTuple = proEqClass->eqClass;
	finalEqTuple = finalEqClass->eqClass;

	minIndex = (int *) malloc(sizeof(int)*total);
	memset(minIndex, 0, sizeof(int)*total);
	noOfMin = 0;

	minDist = -1;
	tempIndex = 0;
	for (i = 0; i < noOfProEqTuple; i++)
	{
		if (((dist[tempIndex] < minDist)|| (minDist == -1)) && ( i != targetIndex))
		{
			minDist = dist[tempIndex];
		}
		tempIndex++;
	}
	for (i = 0; i < noOfFinalEqTuple ; i++)
	{
		if (((dist[tempIndex] < minDist))|| (minDist == -1))
		{
			minDist = dist[tempIndex];
		}
		tempIndex++;
	}

	// locate which entries has the min. dist
	tempIndex = 0;
	for (i = 0; i < noOfProEqTuple; i++)
	{
		if ((dist[tempIndex] == minDist) && (i != targetIndex))
		{
			minIndex[noOfMin] = tempIndex;
			noOfMin++;
		}
		tempIndex++;
	}
	for (i = 0; i < noOfFinalEqTuple; i++)
	{
		if (dist[tempIndex] == minDist)
		{
			minIndex[noOfMin] = tempIndex;
			noOfMin++;
		}
		tempIndex++;
	}

	eqTuple1 = proEqTuple[targetIndex];

	// calculate the attribute list
	attrList = (AttributeList *) malloc(sizeof(AttributeList)*noOfMin);
	memset(attrList, 0, sizeof(AttributeList)*noOfMin);
	for (i = 0; i < noOfMin; i++)
	{
		index = minIndex[i];

		if (index < noOfProEqTuple)
		{
			eqTuple2 = proEqTuple[index];
		}
		else
		{
			eqTuple2 = finalEqTuple[index-noOfProEqTuple];
		}

		newAttrList = kDist_getAttributeList(noOfDim, eqTuple1, eqTuple2);

		memcpy(&(attrList[i]), newAttrList, sizeof(AttributeList));

		// free the newAttr. list
		free(newAttrList);
	}

	// among a number of the attribute list
	// find the smallest size
	minNoOfAttr = attrList[0].noOfAttr;		//as there is at least one elt
	for (i = 1; i < noOfMin; i++)
	{
		noOfAttr = attrList[i].noOfAttr;
		if (noOfAttr < minNoOfAttr)
		{
			minNoOfAttr = noOfAttr;
		}
	}

	// get the list of elts with the smallest size
	smallIndex = (int *) malloc(sizeof(int)*noOfMin);
	noOfSmall = 0;
	for (i = 0; i < noOfMin; i++)
	{
		noOfAttr = attrList[i].noOfAttr;
		if (noOfAttr == minNoOfAttr)
		{
			smallIndex[noOfSmall] = i;
			noOfSmall++;
		}
	}

	// among the attr. list (with smallest size)
	// choose the attr. with the greast poss value
	for (i = 0; i < noOfSmall; i++)
	{
		indexMin = smallIndex[i];

		for (j = 0; j < attrList[indexMin].noOfAttr; j++)
		{
			// find the attr. no
			attrNo = attrList[indexMin].attr[j];

			//
			noOfPoss = genAllTable->genDimTable[attrNo].gen[0].noOfPoss;
			
			eval += noOfPoss;
		}

		if (i == 0)
		{
			minEval = eval;
			smallestIndex = i;
		}

		if (eval < minEval)
		{
			minEval = eval;
			smallestIndex = i;
		}

	}

	indexMin = smallIndex[smallestIndex];
	index = minIndex[indexMin];

	free(smallIndex);

	// free the attr list
	for (i = 0; i < noOfMin; i++)
	{
		free(attrList[i].attr);
	}
	free(attrList);

	free(minIndex);

	return index;
}



void kDist_randomGeneralization(generalizedAllTable *genAllTable, configure *conf, equivalenceClass *proEqClass, equivalenceClass *finalEqClass)
{
	equivalenceTuple *eqTuple1;
	equivalenceTuple *eqTuple2;

	equivalenceTuple *newEqTuple;

	int K;

	int noOfDim;

	int id1, id2;
	int targetIndex;
	
	int noOfMember1, noOfMember2;

	int stubNoOfMember;
	equivalenceTuple *stubEqTuple;

	equivalenceTuple *tempEqTuple;
	int tempNo;

	int noOfProEqTuple, noOfFinalEqTuple;
	equivalenceTuple **proEqTuple;
	equivalenceTuple **finalEqTuple;

	int indexMin;

	float *dist;

	noOfDim = conf->noOfDim;
	K = conf->K;

	noOfProEqTuple = proEqClass->noOfEqClass;
	noOfFinalEqTuple = finalEqClass->noOfEqClass;

	proEqTuple = proEqClass->eqClass;
	finalEqTuple = finalEqClass->eqClass;

	// find a random no.
	targetIndex = rand()%noOfProEqTuple;
//printf("^^%d\n", targetIndex);
//fprintf(g_debugFilePtr, "^^%d\n", targetIndex);
//if (targetIndex == 1) 
//	targetIndex = 2;

	eqTuple1 = proEqTuple[targetIndex];

	// find eq. tuple 2
	dist = kDist_getPairDistance(targetIndex, finalEqClass, proEqClass, genAllTable, conf);
	indexMin = kDist_getMinDistIndex(targetIndex, proEqClass, finalEqClass, dist, genAllTable, noOfDim);
	free(dist);
	if (indexMin < noOfProEqTuple)
	{
		eqTuple2 = proEqTuple[indexMin];
	}
	else
	{
		eqTuple2 = finalEqTuple[indexMin-noOfProEqTuple];
	}

	// 
	id1 = eqTuple1->id;
	id2 = eqTuple2->id;

//printf("  ************************ Merging (%d, %d) **********************\n", id1, id2);
fprintf(g_debugFilePtr, "  ************************ Merging (%d, %d) **********************\n", id1, id2);

//fprintf(g_debugFilePtr, "        ************** Before Removal ************** (After Merging (%d, %d)....)\n", eqTuple1->id, eqTuple2->id);
//fprintf(g_debugFilePtr, "Printing Dist Storage...\n");
//printf("        ************** Before Removal ************** (After Merging (%d, %d)....)\n", eqTuple1->id, eqTuple2->id);
//printf("Printing Dist Storage...\n");
//printDistStorage(g_debugFilePtr, distStorage);

	noOfMember1 = eqTuple1->noOfMember;
	noOfMember2 = eqTuple2->noOfMember;


	if (noOfMember1+noOfMember2 < 2*conf->K)
	{
		// merge two equivalence classes
		newEqTuple = mergeTwoEquivalenceTuple(noOfDim, genAllTable, eqTuple1, eqTuple2);

		// remove eq. tuple 1
		kDist_removeEqTupleFromProcFinal(proEqClass, finalEqClass, eqTuple1, K);

		// remove eq. tuple 2
		kDist_removeEqTupleFromProcFinal(proEqClass, finalEqClass, eqTuple2, K);

//fprintf(g_debugFilePtr, "        ************** After Removal ************** (After Merging (%d, %d)....)\n", eqTuple1->id, eqTuple2->id);
//fprintf(g_debugFilePtr, "Printing Dist Storage...\n");
//printf("        ************** After Removal ************** (After Merging (%d, %d)....)\n", eqTuple1->id, eqTuple2->id);
//printf("Printing Dist Storage...\n");
//printDistStorage(g_debugFilePtr, distStorage);

		// insert newEqTuple to proEqClass/finalEqClass
		kDist_insertEqTuple2ProcFinal(proEqClass, finalEqClass, newEqTuple, K);

		// free
		//free(newEqTuple);

		// 
		free_eqClass(eqTuple1);
		free_eqClass(eqTuple2);
		free(eqTuple1);
		free(eqTuple2);
	}
	else
	{
		if (noOfMember1 < K)
		{
			// do nothing
		}
		else
		{
			// swap
			tempEqTuple = eqTuple1;
			eqTuple1 = eqTuple2;
			eqTuple2 = tempEqTuple;

			tempNo = noOfMember1;
			noOfMember1 = noOfMember2;
			noOfMember2 = tempNo;
		}

			// split eqTuple2
			stubNoOfMember = K-noOfMember1;
			stubEqTuple = kDist_splitEquivalenceTuple(eqTuple2, stubNoOfMember, noOfDim);

			// merge stubEqTuple to eqTuple1
			newEqTuple = mergeTwoEquivalenceTuple(noOfDim, genAllTable, eqTuple1, stubEqTuple);

			// remove eq. tuple 1
			kDist_removeEqTupleFromProcFinal(proEqClass, finalEqClass, eqTuple1, K);

			// remove eq. tuple 2
			kDist_removeEqTupleFromProcFinal(proEqClass, finalEqClass, eqTuple2, K);

			// insert newEqTuple to proEqClass/finalEqClass
			kDist_insertEqTuple2ProcFinal(proEqClass, finalEqClass, newEqTuple, K);

			// insert eqTuple2 to proEqClass/finalEqClass
			kDist_insertEqTuple2ProcFinal(proEqClass, finalEqClass, eqTuple2, K);

			// free
			//free(newEqTuple);
			//free(stubEqTuple);

			// 
			free_eqClass(eqTuple1);
			//free_eqClass(eqTuple2);
			free(eqTuple1);
			//free(eqTuple2);	
	}
}

