
#include "predicate.h"

predicate *predicate_new()
{
	predicate *returnValue;

	returnValue = (predicate *) malloc(sizeof(predicate));
	memset(returnValue, 0, sizeof(predicate));

	return returnValue;
}

void predicate_insert(predicate *aPredicate, int value)
{
	int noOfValue;

	noOfValue = aPredicate->noOfValue;

	if (noOfValue == 0)
	{
		aPredicate->value = (int *) malloc(sizeof(int));
	}
	else
	{
		aPredicate->value = (int *) realloc(aPredicate->value, sizeof(int)*(noOfValue+1));
		
	}
	aPredicate->value[noOfValue] = value;
	aPredicate->noOfValue++;
}


condition *condition_new()
{
	condition *returnValue;

	returnValue = (condition *) malloc(sizeof(condition));
	memset(returnValue, 0, sizeof(condition));

	return returnValue;
}

void condition_insert(condition *aCondition, predicate **aPredicate, int dimNo)
{
	int noOfPredicate;

	noOfPredicate = aCondition->noOfPredicate;

	if (noOfPredicate == 0)
	{
		aCondition->dimNo = (int *) malloc(sizeof(int));
		aCondition->predicateList = (predicate **) malloc(sizeof(predicate *));
	}
	else
	{
		aCondition->dimNo = (int *) realloc(aCondition->dimNo, sizeof(int)*(noOfPredicate+1));
		aCondition->predicateList = (predicate **) realloc(aCondition->predicateList, sizeof(predicate *)*(noOfPredicate+1));
	}

	aCondition->dimNo[noOfPredicate] = dimNo;
	aCondition->predicateList[noOfPredicate] = (*aPredicate);
	aCondition->noOfPredicate++;
}

allCondition *allCondition_new()
{
	allCondition *returnValue;

	returnValue = (allCondition *) malloc(sizeof(allCondition));
	memset(returnValue, 0, sizeof(allCondition));

	return returnValue;
}

void allCondition_insert(allCondition *aAllCondition, condition **aCondition)
{
	int noOfCondition;

	noOfCondition = aAllCondition->noOfCondition;

	if (noOfCondition == 0)
	{
		aAllCondition->conditionList = (condition **) malloc(sizeof(condition *));
	}
	else
	{
		aAllCondition->conditionList = (condition **) realloc(aAllCondition->conditionList, sizeof(condition *)*(noOfCondition+1));
	}

	aAllCondition->conditionList[noOfCondition] = (*aCondition);
	aAllCondition->noOfCondition++;
}

// values may appear the same 
predicate *predicate_generatePredicate(int b, int noOfPossValue)
{
	predicate *returnValue;
	int i;
	int count;
	int randNo;

	returnValue = predicate_new();

	if (b > noOfPossValue)
	{
		b = noOfPossValue;
	}

	count = 0;
	for (i = 0; i < b; i++)
	{
		randNo = rand()%noOfPossValue;
		predicate_insert(returnValue, randNo);
	}
	return returnValue;
}

// s: selectivity
condition *predicate_generateCondition(int queryNoOfDim, int noOfQID, double s, int *dimSize)
{
	condition *returnValue;

	int i, j;
	int randNo;
	int count;
	int *tempNo;

	int isExist;

	int b;
	int noOfPossValue;

	predicate *aPredicate;

	int dimNo;

	if (queryNoOfDim > noOfQID)
	{
		queryNoOfDim = noOfQID;
	}

	returnValue = condition_new();

	count = 0;
	tempNo = (int *) malloc(sizeof(int)*queryNoOfDim);
	memset(tempNo, 0, sizeof(int)*queryNoOfDim);

	// find dim no.
	for (i = 0; i < queryNoOfDim; i++)
	{
		isExist = 1;
		while (isExist == 1)
		{
			randNo = rand()%noOfQID;
			// isexist?
			isExist = 0;
			for (j = 0; j < count; j++)
			{
				if (randNo == tempNo[j])
				{
					isExist = 1;
					break;
				}
			}

			if (isExist == 0)
			{
				tempNo[count] = randNo;
				count++;
				break;
			}
		}
	}

	// 
	for (i = 0; i < queryNoOfDim; i++)
	{
		dimNo = tempNo[i];

		noOfPossValue = dimSize[dimNo];
		b = ceiling(noOfPossValue*pow(s, 1.0/(queryNoOfDim+1)));
		aPredicate = predicate_generatePredicate(b, noOfPossValue);
		
		condition_insert(returnValue, &aPredicate, dimNo);
	}

	// for the sen. tuples
	// assume the last dim is sen.
	dimNo = noOfQID;

	noOfPossValue = dimSize[dimNo];
	b = ceiling(noOfPossValue*pow(s, 1.0/(queryNoOfDim+1)));
	aPredicate = predicate_generatePredicate(b, noOfPossValue);
	condition_insert(returnValue, &aPredicate, dimNo);

	free(tempNo);

	return returnValue;
}

allCondition *predicate_generateAllCondition(int noOfCondition, int queryNoOfDim, int noOfQID, double s, int *dimSize)
{
	allCondition *returnValue;
	int i;
	condition *aCondition;

	returnValue = allCondition_new();
	for (i = 0; i < noOfCondition; i++)
	{
		aCondition = predicate_generateCondition(queryNoOfDim, noOfQID, s, dimSize);
		allCondition_insert(returnValue, &aCondition);
	}

	return returnValue;
}


